# -*- coding: utf-8 -*-
from marshmallow.exceptions import ValidationError
import re


class BaseValidator(object):
    def __init__(self, validator=None, error_massage="invalid value"):
        self.error_massage = error_massage
        self.validator = validator

    def __call__(self, value):
        if self.validator and not self.validator(value):
            raise ValidationError(self.error_massage)
        return True


def not_none_and_empty(value):
    if value:
        return True
    else:
        raise ValidationError("value cant be none and empty")


class regex(BaseValidator):
    def __init__(self, regex_str=None, error_massage="invalid value"):
        self.error_massage = error_massage
        self.validator = lambda v: re.match(regex_str, v) is not None


class range(BaseValidator):
    def __init__(self, min=None, max=None):
        self.error_massage = f"the value not in " \
                             f"range ({min if min is not None else ''},{max if max is not None else ''})"
        self.validator = lambda v: not ((min is not None and v < min) or (max is not None and v > max))



# -*- coding: utf-8 -*-
from consul import Consul as BaseConsul
from consul.base import CB, Check
from uuid import uuid4 as uuid
import random
import json
import os


def _is_node_health(node):
    checks = node["Checks"]
    return len(list(filter(lambda c: c["Status"] != "passing", checks))) == 0


def random_balance(health_nodes):
    return random.sample(health_nodes, 1)[0]


def _rr_balance_func():
    Next_RR_Balance_Count = 0

    def balance(health_nodes):
        nonlocal Next_RR_Balance_Count
        if Next_RR_Balance_Count >= len(health_nodes):
            Next_RR_Balance_Count = 0
        rel = health_nodes[Next_RR_Balance_Count]
        Next_RR_Balance_Count += 1
        return rel

    return balance


rr_balance = _rr_balance_func()


class Consul(BaseConsul):
    def service_register(
            self,
            name,
            service_id=None,
            address=None,
            port=None,
            tags=None,
            check=None,
            token=None,
            meta=None,
            # *deprecated* use check parameter
            script=None,
            interval=None,
            ttl=None,
            http=None,
            timeout=None,
            enable_tag_override=False,
            connect=None):

        payload = {'name': name}

        if enable_tag_override:
            payload['enabletagoverride'] = enable_tag_override
        if service_id:
            payload['id'] = service_id
        if address:
            payload['address'] = address
        if port:
            payload['port'] = port
        if tags:
            payload['tags'] = tags

        if meta:
            payload['meta'] = meta

        if connect:
            payload['connect'] = connect

        if check:
            payload['check'] = check

        else:
            payload.update(Check._compat(
                script=script,
                interval=interval,
                ttl=ttl,
                http=http,
                timeout=timeout))

        params = {}
        token = token or self.token
        if token:
            params['token'] = token

        return self.http.put(
            CB.bool(),
            '/v1/agent/service/register',
            params=params,
            data=json.dumps(payload))

    def service_deregister(self, service_id):
        return self.http.put(
            CB.bool(), '/v1/agent/service/deregister/%s' % service_id)


class ConsulRegister(object):
    _service_id = ""

    def __init__(self, app):
        host = app.config.get("CONSUL_HOST", "127.0.0.1")
        port = app.config.get("CONSUL_PORT", "8500")
        self.name = app.config.get("APP_NAME")
        self.service_info = app.config.get("SERVICE_INFO")
        self.service_info["service_id"] = self.service_id
        print(self.service_id)
        # self.consul = BaseConsul(host=host, port=port)
        self.consul = Consul(host=host, port=port)

        if self.name:
            print("register service...")
            # self.consul.agent.service.register()
            self.consul.service_register(name=self.name, **self.service_info)

    @property
    def service_id(self):
        if not self._service_id:
            if self.service_info and self.service_info.get("service_id"):
                self._service_id = self.service_info.get("service_id")
            elif os.path.exists("service_id"):
                with open("service_id", "r", encoding="utf-8") as f:
                    self._service_id = f.readline()
            else:
                self._service_id = self.name + '-' + uuid().hex
                with open("service_id", "w", encoding="utf-8") as f:
                    f.write(self._service_id)
        return self._service_id

    def get_service(self, name):
        return self.consul.health.service(name)

    def get_health_service_node_by_balance(self, name, balance=random_balance):
        nodes = self.get_service(name)[1]
        health_nodes = list(filter(lambda n: _is_node_health(n), nodes))
        return balance(health_nodes) if health_nodes else None

    def unregister(self):
        print(self.service_id)
        print("unregister service...")
        self.consul.service_deregister(self.service_id)

# def consul_unregister(app):
#     if app:
#         host = app.config.get("CONSUL_HOST", "127.0.0.1")
#         port = app.config.get("CONSUL_PORT", "8500")
#         name = app.config.get("APP_NAME")
#         service_info = app.config.get("SERVICE_INFO")
#         service_id = service_info.get("service_id") if service_info and service_info.get("service_id") else name
#         consul = Consul(host=host, port=port)
#         print(service_id)
#         print(consul.service_deregister(service_id))



# -*- coding: utf-8 -*-
import logging

nameToLevel = {
    'CRITICAL': logging.CRITICAL,
    'FATAL': logging.FATAL,
    'ERROR': logging.ERROR,
    'WARN': logging.WARNING,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG,
    'NOTSET': logging.NOTSET,
}


def init_logger(app):

    log_format = app.config.get("LOGGER_FORMAT",
                                "%(asctime)s - %(levelname)s - %(filename)s - %(funcName)s - %(message)s")
    log_filename = app.config.get("LOGGER_FILENAME", "app.log")
    log_level = nameToLevel.get(app.config.get("LOGGER_LEVEL", "NOTSET"), logging.INFO)
    handler = logging.FileHandler(log_filename, encoding='UTF-8')
    handler.setLevel(log_level)
    logging_format = logging.Formatter(log_format)
    handler.setFormatter(logging_format)
    app.logger.root.level = log_level
    app.logger.addHandler(handler)



# -*- coding: utf-8 -*-
from requests.api import request
from .response import ErrorResponse


class ApiRequests(object):
    consul = None

    def __init__(self, consul=None, set_default=True):
        if consul:
            self.consul = consul
            if set_default:
                ApiRequests.consul = consul

    def _base_request(self, method, sys_name, func_name, model_name="default", prefix="", protocol="http", **kwargs):
        node = self.consul.get_health_service_node_by_balance(sys_name)
        if node:
            host = node["Service"]["Address"]
            port = node["Service"]["Port"]
            uri = [prefix if prefix and prefix != "/" else ""]
            if model_name and model_name != "default":
                uri.append(model_name.lower())
            uri.append(func_name)
            url = "{protocol}://{host}:{port}{uri}".format(protocol=protocol, host=host, port=port, uri="/".join(uri))
            resp = request(method, url, **kwargs)

            try:
                rel = resp.json()
                error = rel.get("error")
                if error:
                    error["code"] = "request error: " + error["code"]
                    error["message"] = f"request [{method}] - {sys_name} - {'/'.join(uri)} ({error['message']})"
                return rel.get("data"), ErrorResponse.convert_by_dict(rel.get("error"), resp.status_code)
            except Exception:
                raise Exception("request {0}-{1}-{2} fail".format(sys_name, model_name, func_name), resp.text)
        else:
            raise Exception("[{0}]:not find alive system".format(sys_name))

    def get(self, sys_name, func_name, model_name="default", prefix="", protocol="http", params=None, **kwargs):
        return self._base_request("get", sys_name, func_name, model_name, prefix, protocol, params=params, **kwargs)

    def post(self, sys_name, func_name, model_name="default", prefix="", protocol="http", params=None, **kwargs):
        return self._base_request('post', sys_name, func_name, model_name, prefix, protocol, json=params, **kwargs)

    def put(self, sys_name, func_name, model_name="default", prefix="", protocol="http", params=None, **kwargs):
        return self._base_request('put', sys_name, func_name, model_name, prefix, protocol, json=params, **kwargs)

    def delete(self, sys_name, func_name, model_name="default", prefix="", protocol="http", **kwargs):
        return self._base_request("delete", sys_name, func_name, model_name, prefix, protocol, **kwargs)


def api_get(sys_name, func_name, model_name="default", prefix="", protocol="http", params=None, **kwargs):
    return ApiRequests().get(sys_name, func_name, model_name, prefix, protocol, params, **kwargs)


def api_post(sys_name, func_name, model_name="default", prefix="", protocol="http", params=None, **kwargs):
    return ApiRequests().post(sys_name, func_name, model_name, prefix, protocol, params, **kwargs)


def api_put(sys_name, func_name, model_name="default", prefix="", protocol="http", params=None, **kwargs):
    return ApiRequests().put(sys_name, func_name, model_name, prefix, protocol, params, **kwargs)


def api_delete(sys_name, func_name, model_name="default", prefix="", protocol="http", **kwargs):
    return ApiRequests().delete(sys_name, func_name, model_name, prefix, protocol, **kwargs)



# -*- coding: utf-8 -*-
from flask import jsonify
from werkzeug.exceptions import HTTPException


def response_factory(data=None, error=None):
    if error and isinstance(error, ErrorResponse):
        return jsonify({
            "error": {
                "code": error.code,
                "message": error.message,
                "detail": error.detail
            },
            "data": data
        }), error.status_code
    elif error and isinstance(error, Exception):
        return jsonify({
            "error": {
                "code": "error",
                "message": str(error),
                "detail": None
            },
            "data": data
        })
    else:
        return jsonify({
            "error": None,
            "data": data
        }), 200


class ErrorResponse(Exception):
    def __init__(self, code, message="", status_code=400, error_detail=None):
        self.status_code = status_code
        self.message = message
        self.code = code
        self.detail = error_detail

    @staticmethod
    def convert_by_http(ex):
        if isinstance(ex, HTTPException):
            return ErrorResponse(ex.name, ex.description, ex.code)
        return None

    @staticmethod
    def convert_by_dict(err, status_code=400):
        if err:
            return ErrorResponse(code=err.get("code"), message=err.get("message"), error_detail=err.get("detail"),
                                 status_code=status_code)
        return None

    def __str__(self):
        return "ErrorResponse({},{},{}) {}".format(self.code,self.message,self.detail,self.status_code)


def page_response(data, total):
    return {
        "data": data,
        "total": total
    }



# -*- coding: utf-8 -*-
from werkzeug import exceptions

from .default_decorator import default_decorator, default_error
from .swagger import SwaggerGenerator
from .log import init_logger
from .consul import ConsulRegister
from .api_requests import ApiRequests
import atexit


class Hsrpc(object):

    def __init__(self, app=None, prefix="", decorator=default_decorator, catch_error=True):
        """"
        构造函数
        """
        self.prefix = prefix
        self.catch_error = catch_error
        self.decorator = decorator
        self.app = app

        self.functions = {}

        self.swagger = SwaggerGenerator(app)
        self.logger = None
        self.consul = None
        self.apirequests = None
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """
        初始化应用
        :param app:
        :return:
        """
        if app:
            self._register_view(app)
            # 注册错误捕获方法
            if self.catch_error:
                app.register_error_handler(exceptions.NotFound, default_error)
                app.register_error_handler(exceptions.MethodNotAllowed, default_error)
                app.register_error_handler(exceptions.InternalServerError, default_error)
                app.register_error_handler(exceptions.BadRequest, default_error)

            self.swagger.init(app)
            init_logger(app)
            self.logger = app.logger
            self.consul = ConsulRegister(app)
            self.apirequests = ApiRequests(self.consul)

            def unregister():
                self.consul.unregister()
                pass

            # 注册注销服务方法
            if app.config.get("HSRPC_AUTO_UNREGISTER", False):
                atexit.register(unregister)

    def register(self, func, func_name=None, model_name=None, use_decorator=True, schema=None, description="",
                 tags=None, **kwargs):
        """
        注册方法
        :param tags:
        :param schema:
        :param description:
        :param use_decorator:
        :param func:
        :param func_name:
        :param model_name:
        :param kwargs:
        :return:
        """

        # 生成路径
        name = func_name if func_name else func.__name__
        if use_decorator and self.decorator:
            view = self.decorator(func)
        else:
            view = func
        if model_name:
            path = self.prefix + "/" + model_name + "/" + name
        else:
            path = self.prefix + "/" + name
        path = path.lower()
        # 生成文档
        api_tags = tags if tags else []
        if model_name:
            api_tags.append(model_name)
        elif not api_tags:
            api_tags.append("Default")

        self.swagger.add_path(path, schema, kwargs.get("methods"), description, api_tags)

        function_info = {
            "func": view,
            "endpoint": model_name + "-" + name if model_name else name,
            "args": kwargs
        }
        self.functions[path] = function_info
        if self.app:
            self.app.add_url_rule(path, endpoint=function_info["endpoint"], view_func=function_info["func"],
                                  **function_info["args"])

    def _register_view(self, app):
        for path, func in self.functions.items():
            app.add_url_rule(path, endpoint=func["endpoint"], view_func=func["func"], **func["args"])

    def route(self, func_name=None, model_name=None, use_decorator=True, schema=None, description="", methods=["POST"],
              tags=None, **options):
        """
        注册路由装饰器
        :param tags:
        :param methods:
        :param use_decorator:
        :param schema:
        :param description:
        :param model_name:
        :param func_name:
        :param use_default_decorator:
        :param options: the options to be forwarded to the underlying
                        :class:`~werkzeug.routing.Rule` object.  A change
                        to Werkzeug is handling of method options.  methods
                        is a list of methods this rule should be limited
                        to (``GET``, ``POST`` etc.).  By default a rule
                        just listens for ``GET`` (and implicitly ``HEAD``).
                        Starting with Flask 0.6, ``OPTIONS`` is implicitly
                        added and handled by the standard request handling.
        """

        def decorator(f):
            options["methods"] = methods if isinstance(methods,list) else [methods]
            self.register(f, func_name, model_name, use_decorator, schema, description, tags, **options)
            return f

        return decorator



# -*- coding: utf-8 -*-
import functools
from .response import response_factory, ErrorResponse
from werkzeug.exceptions import HTTPException
from flask import current_app


def default_decorator(func):
    @functools.wraps(func)
    def decorator(*args, **kwargs):
        try:
            rel = func(*args, **kwargs)
            return make_response(rel)

        except ErrorResponse as ex:
            return response_factory(error=ex)

        except Exception as ex:
            current_app.logger.error(ex)
            print("DEBUG:", current_app.config.get("DEBUG", False))
            if current_app.config.get("DEBUG", False):
                raise ex
            return response_factory(error=ex)

    return decorator


def make_response(rel):
    if isinstance(rel, tuple):
        resp = rel[0]
        error = rel[1]
    elif isinstance(rel, ErrorResponse):
        resp = None
        error = rel
    else:
        resp = rel
        error = None
    if resp is None or isinstance(resp, dict) or isinstance(resp, list) or isinstance(resp, tuple):
        return response_factory(data=resp, error=error)
    else:
        return rel


def default_error(ex):
    if isinstance(ex, HTTPException):
        return response_factory(error=ErrorResponse.convert_by_http(ex))
    else:
        return response_factory(error=ErrorResponse("fail", str(ex)))



# -*- coding: utf-8 -*-
from flask import request
from marshmallow import Schema
from ..response import ErrorResponse
from . import validators
from . import fields


class ArgsDict(dict):
    def __getattr__(self, key):
        return self.get(key)

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, key):
        del self[key]


class BaseSchema(Schema):
    def get_args(self):
        method = request.method.upper()
        content_type = request.content_type.lower() if request.content_type else None

        if content_type and "x-www-form-urlencoded" in content_type:
            args = request.args
        elif content_type and "json" in content_type:
            args = request.json
        elif method == "POST" or method == "PUT":
            args = request.form
        else:
            args = request.args

        rel, err = self.load(args)
        if err:
            raise ErrorResponse("args error", "请求参数错误", error_detail=err)
        return ArgsDict(rel)


class PaginationSchema(BaseSchema):
    offset = fields.Integer(missing=0)
    limit = fields.Integer(missing=20)



# -*- coding: utf-8 -*-
import re
import setuptools


def find_version(fname):
    """Attempts to find the version number in the file names fname.
    Raises RuntimeError if not found.
    """
    version = ''
    with open(fname, 'r') as fp:
        reg = re.compile(r'__version__ = [\'"]([^\'"]*)[\'"]')
        for line in fp:
            m = reg.match(line)
            if m:
                version = m.group(1)
                break
    if not version:
        raise RuntimeError('Cannot find version information')
    return version


__version__ = find_version('flask_hsrpc/__version__.py')


setuptools.setup(
    name="flask-hsrpc",
    version=__version__,
    url="https://github.com/AloneFire/flask-hsrpc.git",

    author="alonefire",
    author_email="alonefire@foxmail.com",

    description="flask hsrpc plugin",
    long_description='flask hsrpc plugin',

    packages=setuptools.find_packages(),

    install_requires=[
        "Flask>=1.0.2",
        "marshmallow>=2.15.1",
        "python-consul>=0.7.2",
        "requests>=2.20.0"
    ],

    classifiers=[
        'Programming Language :: Python :: 3.6',
    ],
)



# -*- coding: utf-8 -*-
class SwaggerGenerator(object):
    def __init__(self, app, title="", version="", description=""):
        self.app = app
        self.desc = {
            "swagger": "2.0",
            "info": {
                "title": title if title else app.config.get("APP_NAME", "API") if app else "API",
                "description": description if description else app.config.get("APP_DESC", "") if app else "",
                "version": version if version else app.config.get("APP_VERSION", "1.0.0") if app else "1.0.0"
            },
            "paths": {}
        }

    def init(self, app, title="", version="", description=""):
        self.app = app
        self.desc["info"].update({
            "title": title if title else app.config.get("APP_NAME", "API") if app else "API",
            "version": version if version else app.config.get("APP_VERSION", "1.0.0") if app else "1.0.0",
            "description": description if description else app.config.get("APP_DESC", "") if app else ""
        })

    def add_path(self, path, schema=None, methods=None, description="", tags=[]):
        methods = methods if methods else ["GET"]
        path_desc = {}
        for method in methods:
            path_desc[method.lower()] = {
                "summary": description,
                "description": description,
                "tags": tags,
                "parameters": [{
                    "name": field.load_from if field.load_from else key,
                    "type": field.__type_name__() if hasattr(field, "__type_name__") else field.__class__.__name__,
                    "required": field.required,
                    "in": ["path"] if method in ("GET") else ["json"],
                    "description": field.metadata.get("help", "") if field.metadata else ""
                } for key, field in schema._declared_fields.items()] if schema else []
            }

        if not self.desc["paths"].get(path):
            self.desc["paths"][path] = {}

        self.desc["paths"][path].update(path_desc)

    def to_dict(self):
        return self.desc



# -*- coding: utf-8 -*-
from marshmallow.fields import *
from marshmallow.exceptions import ValidationError
from marshmallow.fields import List as BaseList, Number as BaseNumber
from marshmallow import utils
import enum
from datetime import datetime


class Enum(Field):
    def __type_name__(self):
        enum_item = [f"{str(v.value)} <{k}>" for k, v in self.enum_obj.__members__.items()]
        return f'Enum({" , ".join(enum_item)})'

    def __init__(self, enum_obj, **kwargs):
        super().__init__(**kwargs)
        if issubclass(enum_obj, enum.Enum):
            self.enum_obj = enum_obj
        else:
            raise ValueError("The type is not Enum")

    def _serialize(self, value, attr, obj):
        if value is None:
            return None
        return value.value

    def _deserialize(self, value, attr, data):
        if not self.required and (value is None or value == ''):
            return self.missing
        try:
            rel_value = value
            if isinstance(rel_value, str):
                if rel_value.isdigit():
                    rel_value = int(rel_value)
            return self.enum_obj(rel_value)
        except ValueError as ex:
            raise ValidationError(str(ex))


class List(BaseList):
    def __type_name__(self):
        field = self.container
        return f'List({field.__type_name__() if hasattr(field, "__type_name__") else field.__class__.__name__})'

    def _deserialize(self, value, attr, data):
        if not utils.is_collection(value):
            value = dict(data).get(attr)
            if not utils.is_collection(value):
                try:
                    result = self.container.deserialize(value)
                    return [result]
                except ValidationError:
                    self.fail('invalid')

        result = []
        errors = {}
        for idx, each in enumerate(value):
            try:
                result.append(self.container.deserialize(each))
            except ValidationError as e:
                result.append(e.data)
                errors.update({idx: e.messages})

        if errors:
            raise ValidationError(errors, data=result)

        return result


class Number(BaseNumber):
    def _deserialize(self, value, attr, data):
        if not self.required and (value is None or value == ''):
            return self.missing
        return self._validated(value)


class DateTime(Field):
    def __type_name__(self):
        return f'DateTime(\"{self.date_format}\")'

    def __init__(self, date_format="%Y-%m-%d %H:%M:%S", **kwargs):
        super().__init__(**kwargs)
        self.date_format = date_format

    def _deserialize(self, value, attr, data):
        if not self.required and (value is None or value == ''):
            return self.missing

        local_datetime = datetime.strptime(value, self.date_format)
        return local_datetime
